
/*
  Copyright 1995 The University of Rhode Island and The Massachusetts
  Institute of Technology

  Portions of this software were developed by the Graduate School of
  Oceanography (GSO) at the University of Rhode Island (URI) in collaboration
  with The Massachusetts Institute of Technology (MIT).

  Access and use of this software shall impose the following obligations and
  understandings on the user. The user is granted the right, without any fee
  or cost, to use, copy, modify, alter, enhance and distribute this software,
  and any derivative works thereof, and its supporting documentation for any
  purpose whatsoever, provided that this entire notice appears in all copies
  of the software, derivative works and supporting documentation.  Further,
  the user agrees to credit URI/MIT in any publications that result from the
  use of this software or in any product that includes this software. The
  names URI, MIT and/or GSO, however, may not be used in any advertising or
  publicity to endorse or promote any products or commercial entity unless
  specific written permission is obtained from URI/MIT. The user also
  understands that URI/MIT is not obligated to provide the user with any
  support, consulting, training or assistance of any kind with regard to the
  use, operation and performance of this software nor to provide the user
  with any updates, revisions, new versions or "bug fixes".

  THIS SOFTWARE IS PROVIDED BY URI/MIT "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
  EVENT SHALL URI/MIT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
  DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE ACCESS, USE OR PERFORMANCE
  OF THIS SOFTWARE.
*/

// This file contains functions which read the variables and their description
// from a netcdf file and build the in-memeory DDS. These functions form the
// core of the server-side software necessary to extract the DDS from a
// netcdf data file.
//
// It also contains test code which will print the in-memeory DDS to
// stdout. 
//
// ReZa 10/20/94

// $Log: cdfdds.cc,v $
// Revision 1.1  1996/06/10 20:34:19  jimg
// Added.
//
// Revision 1.7.2.1  1995/10/12  17:13:21  jimg
// Moved variable definitions from for loops outside those loops. For
// gcc-2.7.0/gcc-2.6.3 compatibility.
//
// Revision 1.7  1995/07/09  21:34:02  jimg
// Added copyright notice.
//
// Revision 1.6  1995/06/23  15:35:24  jimg
// Added netio.h.
// Fixed some misc problems which may have come about do to the modifications in
// nc_das.cc and nc_dds.cc (they went from CGIs to simple UNIX filters).
//
// Revision 1.5  1995/02/10  04:47:35  reza
// Updated to use type subclasses, NCArray, NCByte, NCInt32, NCGrid....
//
// Revision 1.4  1995/01/31  20:21:34  reza
//  Modified for new type subclassed DAP implementation.
//
// Revision 1.3  1994/12/22  04:51:44  reza
// Updated to use DODS new named dimension capability.
//
// Revision 1.2  1994/11/03  05:22:02  reza
// Added variable names to variable types Grid and Array.
//
// Revision 1.1  1994/10/28  14:34:03  reza
// First version
//

static char rcsid[]={"$Id: cdfdds.cc,v 1.1 1996/06/10 20:34:19 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <iostream.h>

#include "netio.h"
#include "cdf.h"
#include "DDS.h"
#include "CDFInt32.h"
#include "CDFFloat64.h"
#include "CDFByte.h"
#include "CDFArray.h"
#include "CDFGrid.h"
#include "CDFStr.h"

// Used by ErrMsgT

static char Msgt[255];
#ifdef NEVER
extern void ErrMsgT(char *Msgt);

extern char *name_path(char *path);
#endif


// This function returns the appropriate DODS BaseType for the given 
// netCDF data type.
//

BaseType *
Get_bt(char *varName, long cdfType) 
{
  switch (cdfType) {
  case CDF_BYTE:	
	return (NewByte(varName));

  case CDF_CHAR:	
  case CDF_UCHAR:	
	// kpvt kluge	return (NewStr(varName));
	return (NewByte(varName));

  case CDF_INT1:
  case CDF_UINT1:
  case CDF_INT2:
  case CDF_UINT2:	
  case CDF_INT4:
  case CDF_UINT4:	
	return (NewInt32(varName));

  case CDF_REAL4:	
  case CDF_FLOAT:	
  case CDF_REAL8:	
  case CDF_DOUBLE:
  case CDF_EPOCH:	
	return (NewFloat64(varName));

  }
  return (NewStr(varName));
}

// Given a reference to an instance of class DDS and a filename that refers
// to a netcdf file, read the netcdf file and extract all the dimensions of
// each of its variables. Add the variables and their dimensions to the
// instance of DDS.
//
// Returns: false if an error accessing the netcdf file was detected, true
// otherwise. 

bool
read_descriptors(DDS &dds_table, char *fileName)
{
  CDFid cdfid;
  CDFstatus status;
  long numDims, numVars, numAtts, encoding, majority, maxRec;
  long cdfDimSizes[CDF_MAX_DIMS];
  long dimSizes[CDF_MAX_DIMS+1];
  char* dimNames[CDF_MAX_DIMS+1];
  char varName[CDF_VAR_NAME_LEN+1];
  long varType;
  long recVary;
  long dimVarys[CDF_MAX_DIMS];
  long recDimVarys[CDF_MAX_DIMS+1];
  long numElts;
  Array *ar;

  status = CDFopen(fileName, &cdfid);
  if (status != CDF_OK) {
	sprintf (Msgt,"cdf_dds: Could not open file %s",fileName);
	ErrMsgT(Msgt);
	return false;
  }

  // how many variables? how many global attributes? 
  if (CDFinquire(cdfid, &numDims, cdfDimSizes, &encoding, &majority, 
				 &maxRec, &numVars, &numAtts) != CDF_OK) {
	ErrMsgT("cdf_dds: Could not inquire about CDF file");
	return false;
  }

  // All variables in a CDF file have the same number and size of
  // dimensions.  The record dimension is not included in the number
  // of dimensions.  Therefore we add the record dimension to number
  // of dims and set dim names for all dims.  The record dimension is
  // prepended to the other dims

  numDims += 1;
  dimSizes[0] = maxRec+1;
  dimNames[0] = new char [32];
  sprintf(dimNames[0], "RecDim");

  // kpvt kluge
  /** The CDF file which are being served for up to KPVT use string
	variables as labels.  The KPVT cannot handle extra dimensions
	so this kludge is to reuse an existing demension as the string
	length.  The mods are label with "kpvt kluge"

	In future string variables should be passed to dods as actual
	stings and then converted to what ever the must be to satisfy
	the client software.  Char types should be tranlated as
	Strings not as char arrays.
	**/
  int kpvtDim;

  for (int d = 1; d < numDims; ++d) {
	dimSizes[d] = cdfDimSizes[d-1];
	dimNames[d] = new char [32];
	sprintf(dimNames[d], "Dimension_%d", d);

	// kpvt kluge
	kpvtDim = d;
  }


  // dataset name
  dds_table.set_dataset_name(name_path(fileName));

  //add all the variables in this file to DDS table 
  for (int varId = 0; varId < numVars; varId++) {
	if (CDFvarInquire(cdfid, varId,
					  varName, &varType,&numElts, &recVary, dimVarys) == -1) {
	  sprintf (Msgt,"cdf_dds: Could not get variable number %d", varId);
	  ErrMsgT(Msgt);
	  return false;
	}

	// Store the record and dimension variance flags in an array
	recDimVarys[0] = recVary;
	for (int d = 0; d < numDims; ++d) 
	  recDimVarys[d+1] = dimVarys[d];


	BaseType *bt = Get_bt(varName, varType);

	ar = NewArray(varName);
	ar->add_var(bt);
	for (int d = 0; d < numDims; ++d) 
	  if (recDimVarys[d])
		ar->append_dim(dimSizes[d],dimNames[d]);

	// kpvt kluge
	// if char array reuse a dim as strlen
	if (numElts > 1) 
	  ar->append_dim(dimSizes[kpvtDim],dimNames[kpvtDim]);

	dds_table.add_var(ar);
  }

  for (int d = 0; d < numDims; ++d)
	delete [] dimNames[d];
  return true;
}

#ifdef TEST

int
main(int argc, char *argv[])
{
    DDS dds;

    if(!read_descriptors(dds, argv[1]))
	abort();

    dds.print();
}

#endif

